import opentracing
from flask import (Request, _request_ctx_stack as stack)

class FlaskTracer(opentracing.Tracer):
    '''
    Tracer that can trace certain requests to a Flask app.

    @param tracer the OpenTracing tracer implementation to trace requests with
    '''
    def __init__(self, tracer, trace_all_requests=False, app=None, traced_attributes=[]): 
        self._tracer = tracer
        self._trace_all_requests = trace_all_requests
        self._current_spans = {}

        # tracing all requests requires that app != None
        if self._trace_all_requests:
            @app.before_request
            def start_trace():
                self._before_request_fn(traced_attributes)

            @app.after_request
            def end_trace(response):
                self._after_request_fn()
                return response


    def trace(self, *attributes):
        '''
        Function decorator that traces functions

        NOTE: Must be placed after the @app.route decorator

        @param attributes any number of flask.Request attributes
        (strings) to be set as tags on the created span
        '''
        def decorator(f):
            def wrapper(*args, **kwargs):
                if not self._trace_all_requests:
                    self._before_request_fn(list(attributes))
                    r = f(*args, **kwargs)
                    self._after_request_fn()
                    return r
                else:
                    return f(*args, **kwargs)
            wrapper.__name__ = f.__name__
            return wrapper
        return decorator

    def get_span(self, request=None):
        '''
        Returns the span tracing `request`, or the current request if
        `request==None`.

        If there is no such span, get_span returns None.

        @param request the request to get the span from
        '''
        if request is None:
            request = stack.top.request
        return self._current_spans.get(request, None)

    def _before_request_fn(self, attributes):
        request = stack.top.request
        operation_name = request.endpoint
        headers = {}
        for k,v in request.headers:
            headers[k.lower()] = v
        span = None
        try:
            span_ctx = self._tracer.extract(opentracing.Format.HTTP_HEADERS, headers)
            span = self._tracer.start_span(operation_name=operation_name, child_of=span_ctx)
        except (opentracing.InvalidCarrierException, opentracing.SpanContextCorruptedException) as e:
            span = self._tracer.start_span(operation_name=operation_name, tags={"Extract failed": str(e)})
        if span is None:
            span = self._tracer.start_span(operation_name)
        self._current_spans[request] = span
        for attr in attributes:
            if hasattr(request, attr):
                payload = str(getattr(request, attr))
                if payload:
                    span.set_tag(attr, payload)

    def _after_request_fn(self):
        request = stack.top.request
        span = self._current_spans.pop(request)
        if span is not None:
            span.finish()




from flask import (Flask, request)
import opentracing
from flask_opentracing import FlaskTracer

app = Flask(__name__)
test_app = app.test_client()

empty_tracer = opentracing.Tracer()
tracer_all = FlaskTracer(empty_tracer, True, app, ['url'])
tracer = FlaskTracer(empty_tracer)

def flush_spans(tcr):
    for req in tcr._current_spans:
        tcr._current_spans[req].finish()
    tcr._current_spans = {}

@app.route('/test')
def check_test_works():
    return "Success"

@app.route('/another_test')
@tracer.trace('url', 'url_rule')
def decorated_fn():
    return "Success again"

@app.route('/wire')
def send_request():
    span = tracer.get_span()
    headers = {}
    empty_tracer.inject(span, opentracing.Format.TEXT_MAP, headers)   
    rv = test_app.get('/test', headers=headers)
    return str(rv.status_code)

def test_on():
    assert True

def test_span_creation():
    with app.test_request_context('/test'):
        app.preprocess_request()
        assert tracer_all.get_span(request)
        assert not tracer.get_span(request)
        flush_spans(tracer_all)

def test_span_deletion():
    assert not tracer_all._current_spans
    test_app.get('/test')
    assert not tracer_all._current_spans

def test_requests_distinct():
    with app.test_request_context('/test'):
        app.preprocess_request()
    with app.test_request_context('/test'):
        app.preprocess_request()
        second_span = tracer_all._current_spans.pop(request)
        assert second_span
        second_span.finish()
        assert not tracer_all.get_span(request)
    # clear current spans
    flush_spans(tracer_all)

def test_decorator():
    with app.test_request_context('/another_test'):
        app.preprocess_request()
        assert not tracer.get_span(request)
        assert len(tracer_all._current_spans) == 1
    flush_spans(tracer)
    flush_spans(tracer_all)

    test_app.get('/another_test')
    assert not tracer_all._current_spans
    assert not tracer._current_spans

def test_over_wire():
    rv = test_app.get('/wire')
    assert '200' in rv.data




from setuptools import setup

'''
Flask-OpenTracing
-----------------

This extension provides simple integration of OpenTracing in Flask applications.
'''
version = open('VERSION').read()
setup(
    name='Flask-OpenTracing',
    version=version,
    url='http://github.com/opentracing-contrib/python-flask',
    download_url='https://github.com/opentracing-contrib/python-flask/tarball/'+version,
    license='BSD',
    author='Kathy Camenzind',
    author_email='kcamenzind@lightstep.com',
    description='OpenTracing support for Flask applications',
    long_description=open('README.rst').read(),
    packages=['flask_opentracing', 'tests'],
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=[
        'Flask',
        'opentracing>=1.1,<1.2'
    ],
    tests_require=['pytest'],
    setup_requires=['pytest-runner'],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)
